package ibspan.ecap.merchant.shopdecisions.dataadministration;

import ibspan.ecap.merchant.shopdecisions.dataaccess.DataModificationService;
import ibspan.ecap.merchant.shopdecisions.dataaccess.DemandFact;
import ibspan.ecap.merchant.shopdecisions.dataaccess.SalesFact;
import ibspan.ecap.merchant.shopdecisions.dataaccess.SuppliesFact;
import ibspan.ecap.merchant.shopdecisions.forecasting.ForecastDescription;
import ibspan.ecap.shared.ontology.negotiation.DemandData;
import ibspan.ecap.shared.ontology.negotiation.NegotiationLog;
import ibspan.ecap.shared.util.TimeConverter;

import jade.util.Logger;

import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.logging.Level;

/**
 * A module responsible for staging and administering the data in the data warehouse. Singleton.
 * 
 * @author  Micha Drozdowicz
 */
public class DataAdministrationModule {
	
	private static DataAdministrationModule instance = new DataAdministrationModule();	
	
	private Object demandMonitor = new Object();
	private Object salesMonitor = new Object();
	
	private DataModificationService svc;
	private DateProcessor dateProcessor;
	private TimeProcessor timeProcessor;
	private StrategyProcessor strategyProcessor;
	private TemplateProcessor templateProcessor;
	private ProductProcessor productProcessor;
	private SalesFactProcessor salesFactProcessor;
	private WholesalerProcessor wholesalerProcessor;
	private SuppliesFactProcessor suppliesFactProcessor;
	private DemandFactProcessor demandFactProcessor;
	
	protected Logger myLogger = Logger.getMyLogger(this.getClass().getName()); 
		
	/**
	 * Gets the single instance of DataAdministrationModule.
	 * 
	 * @return single instance of DataAdministrationModule
	 */
	public static DataAdministrationModule getInstance()
	{
		return instance;
	}
	
	private DataAdministrationModule()
	{
		 svc = DataModificationService.getInstance();
		 dateProcessor = new DateProcessor(svc);
		 timeProcessor = new TimeProcessor(svc);
		 strategyProcessor = new StrategyProcessor(svc);
		 templateProcessor = new TemplateProcessor(svc);
		 productProcessor = new ProductProcessor(svc);
		 wholesalerProcessor = new WholesalerProcessor(svc);
		 
		 salesFactProcessor = new SalesFactProcessor();
		 suppliesFactProcessor = new SuppliesFactProcessor();
		 demandFactProcessor = new DemandFactProcessor();
	}
	
	public void initDateDimension(Date startDate, int days)
	{
		Calendar cal = new GregorianCalendar();
		cal.setTime(startDate);
		for(int i=0; i<days; i++)
		{
			try {
				dateProcessor.retrieveDateKey(cal.getTime());
			} catch (SQLException e) {
				if(myLogger.isLoggable(Level.WARNING))
					myLogger.log(Level.WARNING, "Failed retrieving date key for: "+new SimpleDateFormat().format(cal.getTime()), e);
			}
			cal.add(Calendar.DATE, 1);
		}
	}
	
	/**
	 * Records negotiation details in the database.
	 * 
	 * @param log the information about the negotiation
	 * 
	 * @throws DataAccessException exception is thrown if an error during accessing the database occurs thus causing the operation to fail.
	 */
	public void recordNegotiationDetails(NegotiationLog log) throws DataAccessException
	{
		myLogger.log(Level.FINE, "Recording nego log at:"+TimeConverter.getInstance().getDate()+"\n"+log);
		recordNegotiationSalesData(log);
		//record negotiation bid data
		//record negotiation transaction data
	}
	
	/**
	 * Records finished transaction details in the database.
	 * 
	 * @param log the information about the finished transaction
	 * 
	 * @throws DataAccessException exception is thrown if an error during accessing the database occurs thus causing the operation to fail.
	 * @throws InvalidOperationException exception is thrown if the negotiation corresponding to the specified transaction hadn't been recorded before.
	 */
	public void recordFinishedTransactionDetails(TransactionLog log) throws DataAccessException, InvalidOperationException
	{
		myLogger.log(Level.FINE, "Recording transaction log at:"+TimeConverter.getInstance().getDate()+"\n"+log);
		recordTransactionSalesData(log);
		//record transaction data to other tables
	}
	
	/**
	 * Records warehouse delivery details.
	 * 
	 * @param log the information about a product delivery to the warehouse.
	 * 
	 * @throws DataAccessException exception is thrown if an error during accessing the database occurs thus causing the operation to fail.
	 */
	public void recordWarehouseDeliveryDetails(DeliveryLog log) throws DataAccessException
	{
		int productKey;
		int orderDateKey;
		int orderTimeKey;
		int deliveryDateKey;
		int deliveryTimeKey;
		int wholesalerKey;
		
		try
		{
			productKey = productProcessor.retrieveProductKey(log.getDeliveredProductId());
			orderDateKey = dateProcessor.retrieveDateKey(log.getOrderTime());
			orderTimeKey = timeProcessor.retrieveTimeKey(log.getOrderTime());
			deliveryDateKey = dateProcessor.retrieveDateKey(log.getDeliveryTime());
			deliveryTimeKey = timeProcessor.retrieveTimeKey(log.getDeliveryTime());
			wholesalerKey = wholesalerProcessor.retrieveWholesalerKey(log.getWholesalerId());
		} catch (SQLException e)
		{
			throw new DataAccessException("Database error occurred while processing the dimensions.", e);
		}
		SuppliesFact fact = suppliesFactProcessor.createSuppliesFact(orderDateKey, orderTimeKey, deliveryDateKey, deliveryTimeKey, productKey, wholesalerKey, log);
		
		try
		{
			svc.insertSuppliesFact(fact);
		} catch (SQLException e)
		{
			throw new DataAccessException("Database error occurred while inserting supplies fact data.", e);
		}		
	}
	
	/**
	 * Records demand data.
	 * 
	 * @param data the demand data
	 * 
	 * @throws DataAccessException exception is thrown if an error during accessing the database occurs thus causing the operation to fail.
	 */
	public void recordDemandData(DemandData data) throws DataAccessException
	{
		int productKey;
		int dateKey;
		
		try
		{
			productKey = productProcessor.retrieveProductKey(data.getGlobalProductId());
			dateKey = dateProcessor.retrieveDateKey(data.getWhen());
		} catch (SQLException e)
		{
			throw new DataAccessException("Database error occurred while processing the dimensions.", e);	
		}
		DemandFact fact;
		
		try
		{
			fact = svc.getDemandFact(dateKey, productKey);
		} catch (SQLException e)
		{
			throw new DataAccessException("Database error occurred while retrieving demand fact data.", e);	
		}
		
		if(fact == null)
		{
			fact = demandFactProcessor.createDemandFact(productKey, dateKey, data);
			try
			{
				svc.insertDemandFact(fact);
			} catch (SQLException e)
			{
				throw new DataAccessException("Database error occurred while inserting demand fact data.", e);	
			}
		}
		else
		{
			demandFactProcessor.updateDemandFact(fact, data);
			try
			{
				svc.updateDemandFact(fact);
			} catch (SQLException e)
			{
				throw new DataAccessException("Database error occurred while updating demand fact data.", e);	
			}
		}
		
				
	}
	
	/**
	 * Records buyer registrations.
	 * 
	 * @param info the info
	 * 
	 * @throws DataAccessException the data access exception
	 */
	public void recordBuyerRegistrations(BuyerRegistrationInfo info) throws DataAccessException
	{
		//record into transaction table
		recordRegistrationDemandData(info);
		
	}
	
	public void recordForecastData(ForecastDescription forecast) throws DataAccessException
	{
		myLogger.log(Level.FINE, "Recording forecast log at:"+TimeConverter.getInstance().getDate()+"\n"+forecast);
		DataModificationService dms = DataModificationService.getInstance();
		DateProcessor dp = new DateProcessor(dms);
		ProductProcessor pp = new ProductProcessor(dms);
		ForecastFactProcessor fp = new ForecastFactProcessor();
		
		Date startDate = forecast.getPreparationTime();
		Calendar cal = new GregorianCalendar();
		cal.setTime(startDate);
		cal.add(Calendar.SECOND, +forecast.getForecastPeriod());
		cal.add(Calendar.DATE, -1);
		Date endDate = cal.getTime();
		
		int startDateKey;
		int endDateKey;
		int productKey;
		try {
			startDateKey = dp.retrieveDateKey(startDate);
			endDateKey = dp.retrieveDateKey(endDate);
			productKey = pp.retrieveProductKey(forecast.getProductId());
		} catch (SQLException e) {
			throw new DataAccessException("Database error occurred while processing the dimensions.", e);
		}
		
		ForecastFact fact = fp.createForecastFact(startDateKey, endDateKey, productKey, forecast);
		try {
			dms.insertForecastFact(fact);
		} catch (SQLException sqlEx) {
			throw new DataAccessException(
					"Database error occurred while inserting forecast fact data.",
					sqlEx);
		}
	}
	
	private void recordRegistrationDemandData(BuyerRegistrationInfo info) throws DataAccessException
	{
		DemandFact fact;
		int dateKey;
		int productKey;
		synchronized (demandMonitor)		//retrieving the sales fact data and updating it in the database must be done atomically
		{
			try
			{
				dateKey = dateProcessor.retrieveDateKey(info.getRegistrationTime());
				productKey = productProcessor.retrieveProductKey(info.getProductId());
				fact = svc.getDemandFact(dateKey, productKey);
				
			} catch (SQLException e)
			{
				throw new DataAccessException(
						"Error occurred while fetching sales fact data from database.",
						e);
			}

			if (fact == null)
			{
				fact = demandFactProcessor.createDemandFact(productKey, dateKey, 1);
				try
				{
					svc.insertDemandFact(fact);
				} catch (SQLException e)
				{
					throw new DataAccessException(
							"Error occurred while inserting demand fact data in the database.",
							e);
					
				}
			}
			else
			{
				demandFactProcessor.updateDemandFact(fact, 1);
				try
				{
					svc.updateDemandFact(fact);
				} catch (SQLException e)
				{
					throw new DataAccessException(
							"Error occurred while updating demand fact data in the database.",
							e);
					
				}
			}
		}		
	}

	private void recordTransactionSalesData(TransactionLog log) throws DataAccessException, InvalidOperationException
	{
		int negotiationId = log.getNegotiationId();
		SalesFact fact;
		synchronized (salesMonitor)		//retrieving the sales fact data and updating it in the database must be done atomically
		{
			try
			{
				fact = svc.getSalesFact(negotiationId);
			} catch (SQLException e)
			{
				throw new DataAccessException(
						"Error occurred while fetching sales fact data from database.",
						e);
			}

			if (fact == null)
				throw new InvalidOperationException(
						"Appropriate negotiation data not found. Cannot record a completed transaction before a negotiation is recorded.");

			salesFactProcessor.updateSalesFact(fact, log);
			try
			{
				svc.updateSalesFact(fact);
			} catch (SQLException e)
			{
				throw new DataAccessException(
						"Error occurred while updating sales fact data in the database.",
						e);
				
			}
		}
	}

	private void recordNegotiationSalesData(NegotiationLog log) throws DataAccessException
	{
		int endDateKey;
		int endTimeKey;
		int strategyKey;
		int templateKey;
		int productKey;	
		
		try
		{
			endDateKey = dateProcessor.retrieveDateKey(log.getNegotiationEnd());
			endTimeKey = timeProcessor.retrieveTimeKey(log.getNegotiationEnd());
			strategyKey = strategyProcessor.retrieveStrategyKey(log.getStrategyId());
			templateKey = templateProcessor.retrieveTemplateKey(log.getTemplateId());
			productKey = productProcessor.retrieveProductKey(log.getProductId());			
		}catch(SQLException sqlEx)
		{
			throw new DataAccessException("Database error occurred while processing the dimensions.", sqlEx);	
		}	
		
		SalesFact fact = salesFactProcessor.createSalesFact(endDateKey, endTimeKey, strategyKey, templateKey, productKey, log);
		
		try
		{
			svc.insertSalesFact(fact);
		}
		catch(SQLException sqlEx)
		{
			throw new DataAccessException("Database error occurred while inserting sales fact data.", sqlEx);	
		}
	}
	
	
	
}
