package ibspan.ecap.merchant.shopdecisions.forecasting;

import jade.util.Logger;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.logging.Level;

import net.sourceforge.openforecast.DataPoint;
import net.sourceforge.openforecast.DataSet;
import net.sourceforge.openforecast.Forecaster;
import net.sourceforge.openforecast.ForecastingModel;
import net.sourceforge.openforecast.models.AbstractTimeBasedModel;
import net.sourceforge.openforecast.models.DoubleExponentialSmoothingModel;
import net.sourceforge.openforecast.models.TripleExponentialSmoothingModel;

/**
 * A class responsible for calculating a sales forecast for a given product.
 * @author  Micha Drozdowicz
 */
public class ForecastEvaluator
{
	//TODO: wrzucic do configa
	private static final int DEF_HORIZON = 60*60*24*1;	//1 day

	private static final float DEF_AMOUNT = 50;

	private static final float DEF_VARIATION = 50;
	
	protected Logger myLogger = Logger.getMyLogger(this.getClass().getName());

	private HorizonEstimator horizonSetter;
	private DataProvider dataProvider;
	
	private int[] horizonLengths = new int[]{1, 2, 3, 7, 14, 30};
	private int[] cycleLengths = new int[]{7, 30, 365};
	private double alphaTolerance = 0.1;
	private double betaTolerance = 0.1;
	private double gammaTolerance = 0.1;
	
	public ForecastEvaluator()
	{
		//TODO: napisa lepszy horizonEstimator
		horizonSetter = new FixedHorizonEstimator(3);
		dataProvider = new DatabaseDataProvider();		
	}

	/**
	 * Prepares a forecast of the specified product starting at the given date. 
	 *
	 * @param productId the id of the product
	 * @param forecastStart the start date of the forecast
	 * @return the description of the sales forecast of the product starting at the given date
	 * @throws DataLoadException when the data needed for forecast could not be loaded.
	 */
	public ForecastDescription prepareForecast(String productId,
			Date forecastStart) throws DataLoadException
	{
		
		//myLogger.setLevel(Logger.FINER);
		myLogger.entering(getClass().getName(), "prepareForecast",
				new Object[] { productId, forecastStart });

		int horizon = horizonSetter.estimateHorizon(productId, forecastStart);
		
		DataSet data;
		try
		{
			 data = dataProvider.getHistoricalData(productId, forecastStart, horizon);
			 if(data.size()<15 && myLogger.isLoggable(Level.FINER))
			 {
				 myLogger.finer("Historical data: "+data.toString());
			 }
				 
		}
		catch(Exception ex)
		{
			myLogger.log(Level.WARNING, "Error while retrieving historical data for forecasting.", ex);
			throw new DataLoadException("Error while retrieving historical data for forecasting.",ex);
		}
		
		if(data.size() < 3)
		{
			return new ForecastDescription(productId, DEF_HORIZON, DEF_AMOUNT, DEF_VARIATION, forecastStart);
		}
		ForecastingModel model = prepareModel(data, horizon);
		myLogger.fine(model.toString());
		DataPoint forecastPoint;
		try {
			forecastPoint = dataProvider.getFuturePoint(productId, forecastStart, horizon, data);
		} catch (Exception e) {
			myLogger.log(Level.WARNING, "Error while generating a future point for forecasting.", e);
			throw new DataLoadException("Error while generating a future point for forecasting.",e);
		}
		myLogger.finer("maxTime:"+((AbstractTimeBasedModel) model).getMaximumTimeValue());
		myLogger.finer("minTime:"+((AbstractTimeBasedModel) model).getMinimumTimeValue());
		myLogger.finer("Time variable:"+((AbstractTimeBasedModel) model).getTimeVariable());
		
		myLogger.finer("Data time variable:"+data.getTimeVariable());
		myLogger.finer("Data size:"+data.size());
		double nextSalesValue = model.forecast(forecastPoint);
		double deviation = model.getMAD();
				
		ForecastDescription result = new ForecastDescription(productId, horizon*60*60*24, (float) nextSalesValue, (float) deviation, forecastStart);
		
		myLogger.exiting(getClass().getName(), "prepareForecast", result);

		return result;
	}

	private ForecastingModel prepareModel(DataSet data, int periodLength)
	{
		int bestPeriodsPerCycle = 0;
		for (int i = 0; i < cycleLengths.length; i++)
		{
			int cycleLength = cycleLengths[i];
			
			int periodsPerCycle = cycleLength / periodLength;
			if(data.size() > 2*periodsPerCycle && periodsPerCycle > bestPeriodsPerCycle)
			{
				bestPeriodsPerCycle = periodsPerCycle;
			}			
		}
		
		if(bestPeriodsPerCycle > 0)
		{
			data.setPeriodsPerYear(bestPeriodsPerCycle);
			myLogger.fine("Periods per cycle = "+bestPeriodsPerCycle);

			return getBestTripleModel(data);
			//return DoubleExponentialSmoothingModel.getBestFitModel(data);
		}
		else
		{
			return getBestDoubleModel(data);
		}
		//return Forecaster.getBestForecast(data);
	}

	private ForecastingModel getBestDoubleModel(DataSet data) {
		double alpha;
		double beta;
		
		DoubleExponentialSmoothingModel bestModel = null;
		DoubleExponentialSmoothingModel model;
		
		for(alpha = 0.05; alpha <= 0.95; alpha += alphaTolerance)
		{
			for(beta = 0.05; beta <= 0.95; beta += betaTolerance)
			{			
				model = new DoubleExponentialSmoothingModel(alpha, beta);
				model.init(data);
				
				if(bestModel == null || bestModel.getMSE() > model.getMSE())
				{
					bestModel = model;
				}			
			}
		}
		return bestModel;
	}

	private ForecastingModel getBestTripleModel(DataSet data) {
		
		//return TripleExponentialSmoothingModel.getBestFitModel(data);
		double alpha;
		double beta;
		double gamma;
		
		TripleExponentialSmoothingModel bestModel = null;
		TripleExponentialSmoothingModel model;
		
		for(alpha = 0.05; alpha <= 0.95; alpha += alphaTolerance)
		{
			for(beta = 0.05; beta <= 0.95; beta += betaTolerance)
			{
				for(gamma = 0.05; gamma <= 0.95; gamma += gammaTolerance)
				{
					model = new TripleExponentialSmoothingModel(alpha, beta, gamma);
					model.init(data);
					
					if(bestModel == null || bestModel.getMSE() > model.getMSE())
					{
						bestModel = model;
					}
				}
			}
		}
		return bestModel;
	}
}
