package ibspan.ecap.merchant.shopdecisions.forecasting;

import ibspan.ecap.shared.util.TimeConverter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * A class managing a list of forecasts scheduled for preparation. A product
 * added for forecast is simply scheduled at a date of the end of the current
 * forecast period.
 * 
 * @author Micha Drozdowicz
 * 
 */
public class ForecastScheduler
{
	private Map<String, ForecastTask> productTasks;

	public ForecastScheduler()
	{
		productTasks = new Hashtable<String, ForecastTask>();
	}

	/**
	 * Schedules the preparation of the next forecast of the product.
	 * 
	 * @param task the forecast task to schedule.
	 * @return The date of the next forecast preparation.
	 */
	public Date scheduleForecast(ForecastTask task)
	{
		String productId = task.getProductId();
		Date currentForecastEnd = task.getPreviousForecastEnd();
		Date forecastStart = evaluateNextForecastDate(productId,
				currentForecastEnd);

		task.setForecastStart(forecastStart);

		ForecastTask previousTask = productTasks.put(productId, task);

		return forecastStart;
	}

	/**
	 * Evaluates a date at which the next forecast of the given product should begin.
	 *
	 * @param productId The global product id of the product to schedule.
	 * @param currentForecastEnd The time the current forecast for the product ends.
	 * @return a date at which the next forecast of the given product should begin.
	 */
	public Date evaluateNextForecastDate(String productId,
			Date currentForecastEnd)
	{
		Date forecastStart = currentForecastEnd;

		Date now = TimeConverter.getInstance().getDate();
		if (currentForecastEnd.before(now))
			forecastStart = now;

		return forecastStart;
	}

	/**
	 * Retrieves a list of task scheduled for execution between the given dates.
	 *
	 * @param fromDate the beginning of the period to which the scheduled dates belong
	 * @param toDate the end of the period to which the scheduled dates belong
	 * @return a list of scheduled tasks
	 */
	public List<ForecastTask> getScheduledTasks(Date fromDate, Date toDate)
	{
		List<ForecastTask> result = new ArrayList<ForecastTask>();

		for (ForecastTask task : productTasks.values())
		{
			if (task.getForecastStart().compareTo(fromDate) >= 0
					&& task.getForecastStart().compareTo(toDate) <= 0)
			{
				result.add(task);
			}
		}
		return result;
	}

	/**
	 * Retrieves a scheduled task of the specified product.
	 *
	 * @param productId the id of the product the forecast task of which is to be returned 
	 * @return a forecast task of the specified product
	 */
	public ForecastTask getScheduledTask(String productId)
	{
		return productTasks.get(productId);
	}

	/**
	 * Removes scheduled forecasts for the specified product.
	 * 
	 * @param productId
	 *            the global id of the product, which forecasts are to be
	 *            cancelled.
	 */
	public void cancelProductForecasts(String productId)
	{
		productTasks.remove(productId);
	}

	/**
	 * Confirms that the forecast result has been received and dispatched and removing it from
	 * the pending schedule.
	 *
	 * @param task the task to confirm
	 */
	public void confirmTaskCompletion(ForecastTask task)
	{
		ForecastTask scheduled = productTasks.get(task.getProductId());

		if (scheduled.getTaskId() == task.getTaskId())
			productTasks.remove(task.getProductId());
	}
}
