package ibspan.ecap.merchant.shopdecisions.forecasting;

import ibspan.ecap.shared.util.PropertiesProvider;
import ibspan.ecap.shared.util.ScheduledExecutorCompletionService;
import ibspan.ecap.shared.util.TimeConverter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.FutureTask;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * A class responsible for running forecast tasks at the specified times.
 * @author  Micha Drozdowicz
 */
public class ScheduledForecastRunner
{

	private class ForecastFuture implements Future<ForecastDescription>
	{
		private Future<ForecastDescription> future;

		public ForecastFuture(Future<ForecastDescription> future)
		{
			this.future = future;
		}

		public boolean cancel(boolean arg0)
		{
			boolean result = future.cancel(arg0);
			removeTask(future);
			return result;
		}

		public ForecastDescription get() throws InterruptedException,
				ExecutionException
		{
			ForecastDescription result = future.get();
			removeTask(future);
			return result;
		}

		public ForecastDescription get(long arg0, TimeUnit arg1)
				throws InterruptedException, ExecutionException,
				TimeoutException
		{
			ForecastDescription result = future.get(arg0, arg1);
			removeTask(future);
			return result;
		}

		public boolean isCancelled()
		{
			return future.isCancelled();
		}

		public boolean isDone()
		{
			return future.isDone();
		}
	}

	private static final int DEF_CORE_THREAD_POOL_SIZE = 2;
	private static final int DEF_MAX_THREAD_POOL_SIZE = 10;
	private static final long DEF_MAX_THREAD_KEEP_ALIVE_TIME = 120L; // in seconds
	
	private static final String CORE_THREAD_POOL_SIZE_PROP_NAME = "forecastRunner.coreThreadPoolSize";
	private static final String MAX_THREAD_POOL_SIZE_PROP_NAME = "forecastRunner.maxThreadPoolSize";
	private static final String MAX_THREAD_KEEP_ALIVE_TIME_PROP_NAME = "forecastRunner.threadKeepAliveTime";

	private ScheduledExecutorCompletionService<ForecastDescription> scheduler;

	private Map<Future<ForecastDescription>, ForecastTask> tasks;

	public ScheduledForecastRunner()
	{		
		PropertiesProvider propsProvider = PropertiesProvider.getInstance();
		
		int coreThreadPoolSize = propsProvider.getIntProperty(CORE_THREAD_POOL_SIZE_PROP_NAME, DEF_CORE_THREAD_POOL_SIZE);
		int maxThreadPoolSize = propsProvider.getIntProperty(MAX_THREAD_POOL_SIZE_PROP_NAME, DEF_MAX_THREAD_POOL_SIZE);
		long threadKeepAliveTime = propsProvider.getLongProperty(MAX_THREAD_KEEP_ALIVE_TIME_PROP_NAME, DEF_MAX_THREAD_KEEP_ALIVE_TIME);
		
		ScheduledThreadPoolExecutor executor = new ScheduledThreadPoolExecutor(
				coreThreadPoolSize);		
		executor.setMaximumPoolSize(maxThreadPoolSize);
		executor.setKeepAliveTime(threadKeepAliveTime, TimeUnit.SECONDS);

		scheduler = new ScheduledExecutorCompletionService<ForecastDescription>(
				executor);

		tasks = new HashMap<Future<ForecastDescription>, ForecastTask>();
	}

	/**
	 * Schedules a {@link ForecastTask} for execution. The task will be run at the time specified
	 * by its forecast start date retrieved with the <code>task.getForecastStart</code> method. 
	 *
	 * @param task the task to be executed
	 * @return the {@link Future} describing the future result of the task.
	 */
	public Future<ForecastDescription> scheduleForecast(ForecastTask task)
	{		
		TimeConverter tc = TimeConverter.getInstance();
		Date now = tc.getDate();
		long ms = tc.convertSpanFromReal(task.getForecastStart().getTime() - now.getTime());

		Future<ForecastDescription> f = scheduler.schedule(task, ms,
				TimeUnit.MILLISECONDS);
		tasks.put(f, task);

		ForecastFuture wrapped = new ForecastFuture(f);
		task.setFuture(wrapped);

		return wrapped;
	}

	/**
	 * Immediately executes the specified task.  
	 *
	 * @param task the task to be executed
	 * @return the {@link Future} describing the future result of the task.
	 */
	public Future<ForecastDescription> executeForecast(ForecastTask task)
	{

		Future<ForecastDescription> f = scheduler.submit(task);
		tasks.put(f, task);

		ForecastFuture wrapped = new ForecastFuture(f);
		task.setFuture(wrapped);

		return wrapped;
	}

	/**
	 * Performs a blocking receive of a finished task. Blocks the execution of the calling thread
	 * until a task completes and the result is available.
	 *
	 * @return a completed forecast task.
	 * @throws InterruptedException when the calling thread has been interrupted while waiting
	 * for the result
	 */
	public ForecastTask takeCompletedTask() throws InterruptedException
	{
		Future<ForecastDescription> f = scheduler.take();
		return removeTask(f);
	}

	/**
	 * Performs a blocking receive of a finished task with a timeout. Blocks the execution of the calling thread
	 * until a task completes and the result is available or until the timeout has elapsed.
	 *
	 * @param millisecondsTimeout the timeout in millis after which the method should return 
	 * regardless of the availability of the result.
	 * @return a completed forecast task or null if the method timed out before receiving 
	 * a result.
	 * @throws InterruptedException when the calling thread has been interrupted while waiting
	 * for the result
	 */
	public ForecastTask pollForecastResult(long millisecondsTimeout)
			throws InterruptedException
	{
		Future<ForecastDescription> f = scheduler.poll(millisecondsTimeout,
				TimeUnit.MILLISECONDS);
		return removeTask(f);
	}

	/**
	 * Performs a non-blocking receive of a finished task. 
	 *
	 * @return a completed forecast task if it was available at the time of calling the method
	 * or null otherwise
	 */
	public ForecastTask pollForecastResult()
	{
		Future<ForecastDescription> f = scheduler.poll();
		return removeTask(f);
	}

	/**
	 * Cancels forecasts for the specified product.
	 * 
	 * @param productId
	 *            the global id of the product, which forecasts are to be
	 *            cancelled.
	 */
	public void cancelProductForecasts(String productId)
	{
		Collection<Future<ForecastDescription>> toRemove = new ArrayList<Future<ForecastDescription>>();

		for (Map.Entry<Future<ForecastDescription>, ForecastTask> entry : tasks
				.entrySet())
		{
			if (entry.getValue().getProductId().equals(productId))
			{
				toRemove.add(entry.getKey());
			}
		}

		for (Future<ForecastDescription> f : toRemove)
		{
			f.cancel(true);
			tasks.remove(f);
		}
	}

	protected ForecastTask removeTask(Future<ForecastDescription> future)
	{
		if (future != null)
			return tasks.remove(future);
		else
			return null;
	}
}
